import numpy as np

deg2rad = np.pi/180
rad2deg = 1 / deg2rad

def hat(x):
    return np.array([[0, -x[2], x[1]], [x[2], 0, -x[0]], [-x[1], x[0], 0]])

def inv_hat(m):
    return np.array([m[2, 1], m[0, 2], m[1, 0]])

# (a, n): a = angle, n = vecteur unitaire (axe de rotation)
def quat2axang(q):
    a = 2 * np.arccos(q[0])
    n = q[1:4] / np.sqrt(1 - q[0] * q[0])
    return a, n

def dcm2axang(R):
    t = np.trace(R)
    n = 0.5 * inv_hat(R - R.T)
    r = np.linalg.norm(n)
    a = np.arctan2(r, (t - 1) / 2)
    return a, n / r


# Quaternions
def quatmat(p):
    m = np.array(
        [
            [p[0], -p[1], -p[2], -p[3]],
            [p[1],  p[0], -p[3],  p[2]], 
            [p[2],  p[3],  p[0], -p[1]], 
            [p[3], -p[2],  p[1],  p[0]]
        ]
    )
    return m

def quatmul(p, q):
    return quatmat(p)@q

def quatnorm(q): 
    return np.dot(q, q)

def quatconj(q):
    return q * np.array([1, -1, -1, -1])

def quatinv(q):
    return quatconj(q) / quatnorm(q)

def vect2quat(u):
    return np.concatenate([[0], u])

def quatrot(q, u):
    v = quatmul(quatinv(q), quatmul(vect2quat(u), q))
    return v[1:]


def axang2quat(a, n):
    return np.concatenate([[np.cos(a / 2)], n * np.sin(a / 2)])
    
def eul2quat(a):
    qr = np.array([np.cos(a[0]/2), np.sin(a[0]/2), 0, 0])
    qp = np.array([np.cos(a[1]/2), 0, np.sin(a[1]/2), 0])
    qy = np.array([np.cos(a[2]/2), 0, 0, np.sin(a[2]/2)])
    return quatmul(qy, quatmul(qp, qr))


# Matrices de Rotation (DCM)
def axang2dcm(a, n):
    dcm = (1 - np.cos(a)) * np.outer(n, n) + np.cos(a) * np.identity(3) - np.sin(a) * hat(n)
    return dcm

def quat2dcm(q):
    w = q[1:4]
    dcm = (2 * np.outer(w, w) + (q[0]**2 - w @ w) * np.identity(3) - 2 * q[0] * hat(w)) / quatnorm(q)
    return dcm

def eul2dcm(a):
    Rx = np.array([[1, 0, 0], [0, np.cos(a[0]), np.sin(a[0])], [0, -np.sin(a[0]), np.cos(a[0])]])
    Ry = np.array([[np.cos(a[1]), 0, -np.sin(a[1])], [0, 1, 0], [np.sin(a[1]), 0, np.cos(a[1])]])
    Rz = np.array([[np.cos(a[2]), np.sin(a[2]), 0], [-np.sin(a[2]), np.cos(a[2]), 0], [0, 0, 1]])               
    return Rx @ Ry @ Rz


# Angles d'Euler (phi, theta, psi) selon axes (x, y, z) : z en premier
def quat2eul(q):
    phi = np.arctan2(2 * (q[0] * q[1] + q[2] * q[3]), q[0]**2 - q[1]**2 - q[2]**2 + q[3]**2)
    theta = -np.arcsin(2 * (q[1] * q[3] - q[0] * q[2]))
    psi = np.arctan2(2 * (q[0] * q[3] + q[1] * q[2]), q[0]**2 + q[1]**2 - q[2]**2 - q[3]**2)
    return np.array([phi, theta, psi])
    
def dcm2eul(R):
    phi = np.arctan2(R[1, 2], R[2, 2])
    theta = -np.arcsin(R[0, 2])
    psi = np.arctan2(R[0, 1], R[0, 0])
    return np.array([phi, theta, psi])
