import numpy as np
import gymnasium as gym
from gymnasium.envs.classic_control import utils
from mavion_model import Mavion
from rotation import *


deg2rad = np.pi / 180

class MavionEnv(gym.Env):

    def __init__(self, mavion=Mavion(), render_mode=None):

        self.tau = 0.2  # seconds between state updates

        # Angle at which to fail the episode
        self.pos_threshold = np.array([2.4, 2.4, 2.4])
        self.ang_threshold = np.array([15, 15, 15]) * deg2rad
        self.quat_threshold = np.array([1, 1, 1, 1])
        self.vel_threshold = np.finfo(np.float32).max * np.ones(3)
        self.rot_threshold = np.finfo(np.float32).max * np.ones(3)

        high = np.concatenate(
            [
                self.pos_threshold * 2,
                # self.ang_threshold * 2,
                self.quat_threshold,
                self.vel_threshold,
                self.rot_threshold
            ],
            dtype=np.float32,
        )
        self.observation_space = gym.spaces.Box(-high, high, dtype=np.float32)
        self.action_space = gym.spaces.Box(-1, 1, shape=(4,), dtype=np.float32)
        self.action_range = np.array([mavion.MAX_ROTOR_SPD, mavion.MAX_ROTOR_SPD, mavion.MAX_FLAP_DEF, mavion.MAX_FLAP_DEF])

        self.state = np.zeros(13)
        self.target = np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        self.obs = self.state - self.target
        self.steps_beyond_terminated = None
        self.mavion = mavion

    def reset(self, seed=None, options=None):
        super().reset(seed=seed)
        self.obs = np.zeros(13)
        low, high = utils.maybe_parse_reset_bounds(options, -0.1, 0.1) 
        self.obs[0:3] = self.np_random.uniform(low=low, high=high, size=(3,))
        self.state = self.target + self.obs
        self.steps_beyond_terminated = None
        return self.obs, {}

    def distance(self, s, t):
        return np.linalg.norm(t[0:3] - s[0:3])

    def cost(self, s, u, t):
        u_trim = self.mavion.trim(t)
        s_trim = self.mavion.trim_state(t)
        J = (s - s_trim).T@Q@(s - s_trim) + (u - u_trim).T@Q@(u - u_trim)
        return J

    def step(self, action):
        x = self.state
        u = action * self.action_range
        w = np.zeros(3)
        new_state = self.mavion.step(x, u, w, self.tau)
        self.state = new_state
        terminated = self.distance(self.state, self.target) > 10
        reward = 1 if not terminated else -10
        self.obs = self.state - self.target
        return self.obs, reward, terminated, False, {}
